"""Root conftest.py file for pytest configuration."""

# pylint: disable=redefined-outer-name

import datetime as dt
import os
import uuid
from unittest import mock

import jwt
from pytest import fixture


@fixture(scope="session")
def endpoint():
    """Return the server URL."""
    return os.environ["DRIFT_MONITOR_URL"]


@fixture(scope="session")
def token(request):
    """Return the server token."""
    if hasattr(request, "param") and request.param:
        return request.param
    now = dt.datetime.now(dt.timezone.utc).timestamp()
    payload = {
        "sub": "1234567890",
        "name": "John Doe",
        "iat": now,
        "exp": now + 10000000,
    }
    token = jwt.encode(payload, "some_key", algorithm="HS256")
    return token


@fixture(scope="session", autouse=True)
def token_mock(token):
    """Patch the access token with a MagicMock."""
    with mock.patch("drift_monitor.utils.access_token") as access_token:
        access_token.return_value = token
        yield access_token


@fixture(scope="module")
def request_mock():
    """Patch requests module with MagicMocks."""
    with mock.patch("drift_monitor.utils.requests") as requests:
        yield requests


@fixture(scope="session")
def experiment(request):
    """Return a new experiment."""
    kwds = request.param if hasattr(request, "param") else {}
    return {
        "id": kwds.get("id", f"{uuid.uuid4()}"),
        "created_ad": dt.datetime.now().isoformat(),
        "name": kwds.get("name", "new_experiment"),
        "description": kwds.get("description", "A new experiment."),
        "public": kwds.get("public", False),
        "permissions": kwds.get("permissions", {}),
    }


@fixture(scope="session")
def drift(request):
    """Return a new drift with parameters from request."""
    default_drift = {
        "id": str(uuid.uuid4()),
        "created_at": dt.datetime.now().isoformat(),
        "model": "model_1",
        "job_status": "Running",
        "concept_drift": {},
        "data_drift": {},
    }
    if hasattr(request, "param"):
        default_drift.update(request.param)
    return default_drift



"""Utility functions for drift monitor."""

import requests

from drift_monitor.config import access_token, settings


def create_experiment(attributes):
    """Create a new experiment on the drift monitor server."""
    response = requests.post(
        url=f"{settings.monitor_url}/experiment",
        headers={"Authorization": f"Bearer {access_token()}"},
        json=attributes,
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return response.json()


def get_experiment(experiment_name):
    """Get an experiment from the drift monitor server."""
    response = requests.get(
        url=f"{settings.monitor_url}/experiment",
        headers={"Authorization": f"Bearer {access_token()}"},
        json={"name": experiment_name},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return response.json()[0] if response.json() else None


def create_drift(experiment, model):
    """Create a drift run on the drift monitor server."""
    exp_route = f"experiment/{experiment['id']}"
    response = requests.post(
        url=f"{settings.monitor_url}/{exp_route}/drift",
        headers={"Authorization": f"Bearer {access_token()}"},
        json={"model": model, "job_status": "Running"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return response.json()


def complete_drift(experiment, drift):
    """Complete a drift run on the drift monitor server."""
    _drift = {k: v for k, v in drift.items() if k not in {"id", "created_at"}}
    exp_route = f"experiment/{experiment['id']}"
    response = requests.put(
        url=f"{settings.monitor_url}/{exp_route}/drift/{drift['id']}",
        headers={"Authorization": f"Bearer {access_token()}"},
        json={**_drift, "job_status": "Completed"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()


def fail_drift(experiment, drift):
    """Fail a drift run on the drift monitor server."""
    _drift = {k: v for k, v in drift.items() if k not in {"id", "created_at"}}
    exp_route = f"experiment/{experiment['id']}"
    response = requests.put(
        url=f"{settings.monitor_url}/{exp_route}/drift/{drift['id']}",
        headers={"Authorization": f"Bearer {access_token()}"},
        json={**_drift, "job_status": "Failed"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()


def register():
    """Registers the token user in the application database."""
    response = requests.post(
        url=f"{settings.monitor_url}/user",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()


def update_email():
    """Update the email of the token user in the application database."""
    response = requests.put(
        url=f"{settings.monitor_url}/user/self",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()



"""Test module for creation of experiments."""

from unittest import mock

import pytest

from drift_monitor import DriftMonitor


@pytest.fixture(scope="module", autouse=True)
def mocks(request_mock, experiment, drift):
    """Mock the requests module."""
    m = mock.MagicMock(json=lambda: [experiment.copy()])
    request_mock.get.return_value = m
    m = mock.MagicMock(json=lambda: drift.copy())
    request_mock.post.return_value = m


@pytest.fixture(scope="module")
def create_drift(request_mock, experiment, drift):
    """Create a drift run on the drift monitor server."""
    with DriftMonitor(experiment["name"], drift["model"]) as monitor:
        monitor.concept(True, {"threshold": 0.5})
        monitor.data(True, {"threshold": 0.5})
        return monitor


@pytest.mark.usefixtures("create_drift")
def test_running_drift(request_mock, endpoint, experiment, token):
    """Test the drift run was created on the server."""
    assert request_mock.post.call_count == 1
    url = f"{endpoint}/api/v1/experiment/{experiment["id"]}/drift"
    assert request_mock.post.call_args[1]["url"] == url
    assert request_mock.post.call_args[1]["headers"] == {
        "Authorization": f"Bearer {token}",
    }
    json = request_mock.post.call_args[1]["json"]
    assert json["job_status"] == "Running"
    assert "concept_drift" not in json
    assert "data_drift" not in json


@pytest.mark.usefixtures("create_drift")
def test_completed_drift(request_mock, endpoint, experiment, token, drift):
    """Test the drift run was completed on the server."""
    assert request_mock.put.call_count == 1
    url = f"{endpoint}/api/v1/experiment/{experiment["id"]}/drift/{drift["id"]}"
    assert request_mock.put.call_args[1]["url"] == url
    assert request_mock.put.call_args[1]["headers"] == {
        "Authorization": f"Bearer {token}",
    }
    json = request_mock.put.call_args[1]["json"]
    assert json["job_status"] == "Completed"
    assert json["concept_drift"] == {"drift": True, "parameters": {"threshold": 0.5}}
    assert json["data_drift"] == {"drift": True, "parameters": {"threshold": 0.5}}
